//
// Created by jordi on 11/02/18.
//

#include <armadillo>
#include "BallHistory.h"

BallHistory::BallHistory() = default;

void BallHistory::addTrackingElement(const HistoryElem &elem) {
    trackingHistory.push_back(elem);
    internalTrackingHistory.push_back(elem);
}

void BallHistory::addBounceElement(const HistoryElem &elem) {
    bounceHistory.push_back(elem);
}

bool BallHistory::wasFullyDetected() {
    return trackingHistory.back().wasFullyDetected();
}

cv::Point2f BallHistory::getLastPoint() {
    return trackingHistory.back().getPoint();
}

void BallHistory::printHistory() {
    for (const auto &elem : trackingHistory) {
        auto point = elem.getPoint();
        std::cout << point.x << ", " << point.y << ", " << elem.getTimestamp() << std::endl;
    }
    std::cout << "Bounce History" << std::endl
              << "--------------" << std::endl;
    for (const auto &elem : bounceHistory) {
        auto point = elem.getPoint();
        std::cout << point.x << ", " << point.y << ", " << elem.getTimestamp() << std::endl;
    }
}

cv::Point2f BallHistory::getBounce() {
    if (internalTrackingHistory.size() < 4 || numOfLastDetections(4) != 4)
        return NULL_POINT;
    auto lastPoints = internalTrackingHistory.rbegin();

    std::vector<float> x_points(4), y_points(4);
    for (int i = 0; i < 4; i++) {
        const auto point = (*lastPoints).getPoint();
        x_points[i] = point.x;
        y_points[i] = point.y;
        lastPoints++;
    }

    auto residual = calculatePolynomialFittingResidual(x_points, y_points);
//    std::cout << residual << std::endl;
    if (residual > RESIDUAL_THRESHOLD) {
        internalTrackingHistory.clear();
        std::vector<float> sub_x(&x_points[1], &x_points[4]);
        std::vector<float> sub_y(&y_points[1], &y_points[4]);
        auto bounceX = (x_points[0] + x_points[1]) / 2;
        return calculateBouncePoint(sub_x, sub_y, bounceX);
    }
    return NULL_POINT;
}

float BallHistory::calculatePolynomialFittingResidual(std::vector<float> x, std::vector<float> y) {
    arma::fmat vandermonde = {
            {1, x[0], x[0] * x[0]},
            {1, x[1], x[1] * x[1]},
            {1, x[2], x[2] * x[2]},
            {1, x[3], x[3] * x[3]}
    };
    arma::fvec y_points(y);
    try {
        arma::fmat Q, R;
        arma::qr_econ(Q, R, vandermonde);
        arma::fvec coefficients = R.i() * (Q.t() * y_points);
//        arma::fvec coefficients = arma::solve(vandermonde, y_points);
        arma::fvec residual = y_points - (vandermonde * coefficients);
//        std::cout << residual << std::endl;
        residual.transform([](float val) { return val * val; });
        return arma::sum(residual);
    } catch (const std::runtime_error &ex) {
        return -1;
    }
}

cv::Point2f BallHistory::calculateBouncePoint(std::vector<float> x, std::vector<float> y, float bounceX) {
    arma::fmat vandermonde = {
            {1, x[0], x[0] * x[0]},
            {1, x[1], x[1] * x[1]},
            {1, x[2], x[2] * x[2]}
    };
    arma::fvec y_points(y);
    try {
        auto coefficients = arma::solve(vandermonde, y_points);
        arma::fvec poly_x = {1, bounceX, bounceX * bounceX};
        float bounceY = arma::dot(poly_x, coefficients);
        return cv::Point2f(bounceX, bounceY);
    } catch (const std::runtime_error &ex) {
        return NULL_POINT;
    }
}

int BallHistory::numOfLastDetections(int n) {
    auto it = internalTrackingHistory.rbegin();
    int result = 0;
    for (int i = 0; i < n; i++) {
        if ((*it).getPoint() != NULL_POINT)
            result++;
    }
    return result;
}

std::list<HistoryElem> BallHistory::mergedHistories() {
    auto merged = std::list<HistoryElem>();
    merged.insert(merged.end(), trackingHistory.begin(), trackingHistory.end());
    merged.insert(merged.end(), bounceHistory.begin(), bounceHistory.end());
    merged.sort();
    return merged;
}
