//
// Created by jordi on 11/02/18.
//

#include <opencv2/core/gpumat.hpp>
#include "Pipeline.h"


using namespace cv;

bool Pipeline::process(const cv::Mat &mat) {
    auto results = this->detectBall(mat);
    auto detectedPoint = std::get<0>(results);
    auto fullyDetected = std::get<1>(results);
    ballHistory.addTrackingElement(HistoryElem(detectedPoint, -1, fullyDetected, false));
    auto bouncePoint = ballHistory.getBounce();
    if (bouncePoint != NULL_POINT) {
        ballHistory.addBounceElement(HistoryElem(bouncePoint, -1, fullyDetected, true));
    }
    return bouncePoint != NULL_POINT;
}

Pipeline::Pipeline() {
    gaussianFilter = cv::cuda::createGaussianFilter(CV_8U, CV_8U, Size(5, 5), 0);
    backgroundSubtractor = cv::cuda::createBackgroundSubtractorMOG2(500, 16, false);
    stream = cv::cuda::Stream::Null();
}

std::tuple<Point2f, bool> Pipeline::detectBall(const cv::Mat &mat) {
    cuda::GpuMat sink(mat), foreground, isolated, converted, filtered;
    cuda::GpuMat colorFiltered, hsvImage, satFilter, valFilter, newMask;
    std::vector<cuda::GpuMat> channels(3);
    cuda::cvtColor(sink, converted, cv::COLOR_BGR2GRAY, 0, stream);
    backgroundSubtractor->apply(converted, foreground, -1, stream);
    int radius = 30;
    auto lastHistory = ballHistory.trackingHistory.back();
    auto previousToLastHistory = (*ballHistory.trackingHistory.rbegin()++);
    auto point = lastHistory.getPoint();
    auto previousPoint = previousToLastHistory.getPoint();
    auto offset = Point2f(0, 0);
    if (point == NULL_POINT && previousPoint != NULL_POINT)
        radius *= 2;
    if (point != NULL_POINT || previousPoint != NULL_POINT) {
        auto x = static_cast<int>(point.x), y = static_cast<int>(point.y);
        int minX = max(x - radius, 0), maxX = min(x + radius, mat.size().width);
        int minY = max(y - radius, 0), maxY = min(y + radius, mat.size().height);
        offset = Point2f(minX, minY);
        auto rangeX = Range(minX, maxX);
        auto rangeY = Range(minY, maxY);
        // Clone is necessary due to a misaligned address eception otherwise for god knows what reason
        auto foregroundMask = foreground(rangeY, rangeX).clone();
        if (mat.channels() == 3 || mat.channels() == 4) {
            cuda::cvtColor(sink(rangeY, rangeX), hsvImage, cv::COLOR_BGR2HSV, 0, stream);
            cuda::split(hsvImage, channels, stream);
            cuda::threshold(channels[1], satFilter, 100, 255, THRESH_BINARY_INV, stream);
            cuda::threshold(channels[2], valFilter, 150, 255, THRESH_BINARY, stream);
            cuda::bitwise_and(valFilter, satFilter, colorFiltered, noArray(), stream);
            cuda::bitwise_and(colorFiltered, foregroundMask, newMask, noArray(), stream);
            converted(rangeY, rangeX).copyTo(isolated, newMask, stream);
        } else {
            converted(rangeY, rangeX).copyTo(isolated, foregroundMask, stream);
        }
    } else {
        if (mat.channels() == 3 || mat.channels() == 4) {
            cuda::cvtColor(sink, hsvImage, cv::COLOR_BGR2HSV, 0, stream);
            cuda::split(hsvImage, channels, stream);
            cuda::threshold(channels[1], satFilter, 100, 255, THRESH_BINARY_INV, stream);
            cuda::threshold(channels[2], valFilter, 150, 255, THRESH_BINARY, stream);
            cuda::bitwise_and(valFilter, satFilter, colorFiltered, foreground, stream);
            cuda::bitwise_and(foreground, colorFiltered, newMask, noArray(), stream);
            converted.copyTo(isolated, newMask, stream);
        } else {
            converted.copyTo(isolated, foreground, stream);
        }
    }
    gaussianFilter->apply(isolated, filtered, stream);
    stream.waitForCompletion();
    cv::Mat final(filtered);
    bool fullyDetected = true;
    auto result = ballDetector.detect(final, true);
    if (result == NULL_POINT && (point != NULL_POINT || previousPoint != NULL_POINT) &&
        (lastHistory.wasFullyDetected())) {
        result = ballDetector.detect(final, false);
        fullyDetected = false;
    }
//    cv::Mat debug(filtered);
//    cv::imshow("Debug", debug);
//    cv::waitKey(1);
    if (result != NULL_POINT) {
        result += offset;
    }
    return std::make_tuple(result, fullyDetected);
}
