#include <iostream>
#include <fstream>
#include <chrono>

#include <opencv2/opencv.hpp>
#include "cxxopts.hpp"

#include "Pipeline.h"


int main(int argc, const char * argv[]) {
    cxxopts::Options options("ball-detector", "This program detects balls and prints its location");
    options.add_options()
            ("i,input", "Input file to process", cxxopts::value<std::string>(), "FILE")
            ("v,video", "Save video with visualization to specified path", cxxopts::value<std::string>(), "FILE")
            ("b,bounces-out", "Save bounces to specified path", cxxopts::value<std::string>(), "FILE")
            ("d,detections-out", "Save ball detections to specified path",
             cxxopts::value<std::string>()->default_value("stdout"), "FILE")
            ("h,help", "Print this help");
    std::string input, video, detections;
    std::streambuf *detectionBuf;
    std::streambuf *bounceBuf;
    std::ofstream detectionOutFile, bounceOutFile;
    bool outputVideo;
    try {
        auto result = options.parse(argc, argv);
        outputVideo = result.count("video") > 0;
        if (result.count("help")) {
            std::cout << options.help() << std::endl;
            exit(0);
        }
        if (!result.count("input")) {
            std::cout << options.help() << std::endl;
            exit(1);
        }
        input = result["input"].as<std::string>();
        if (result.count("detections-out")) {
            detectionOutFile.open(result["detections-out"].as<std::string>());
            detectionBuf = detectionOutFile.rdbuf();
        } else {
            detectionBuf = std::cout.rdbuf();
        }
        if (result.count("bounces-out")) {
            bounceOutFile.open(result["bounces-out"].as<std::string>());
            bounceBuf = bounceOutFile.rdbuf();
        } else {
            bounceBuf = nullptr;
        }
        if (outputVideo) {
            video = result["video"].as<std::string>();
        }
    } catch (const cxxopts::OptionException &e) {
        std::cout << "error parsing options: " << e.what() << std::endl;
        exit(1);
    }
    std::ostream out(detectionBuf);
    std::ostream bounceOut(bounceBuf);
    cv::VideoCapture videoCapture(input);
    Pipeline pipeline;
    cv::Mat frame;
    while (videoCapture.read(frame)) {
        const auto hasBounced = pipeline.process(frame);
        const auto &history = pipeline.ballHistory.trackingHistory.back();
        const auto &point = history.getPoint();
        out << point.x << "," << point.y << "," << history.getTimestamp() << std::endl;
        if (bounceBuf != nullptr && hasBounced) {
            const auto &bounceHistory = pipeline.ballHistory.bounceHistory.back();
            const auto &bouncePoint = bounceHistory.getPoint();
            bounceOut << bouncePoint.x << "," << bouncePoint.y << "," << bounceHistory.getTimestamp() << std::endl;
        }
    }
    if (outputVideo) {
        cv::VideoWriter videoWriter(video, VideoWriter::fourcc('H', '2', '6', '4'), 23.976,
                                    {static_cast<int>(videoCapture.get(cv::CAP_PROP_FRAME_WIDTH)),
                                     static_cast<int>(videoCapture.get(cv::CAP_PROP_FRAME_HEIGHT))});
        videoCapture.set(cv::CAP_PROP_POS_FRAMES, 0);
        auto mergedHistories = pipeline.ballHistory.mergedHistories();
        auto it = mergedHistories.begin();
        auto begin = mergedHistories.begin();
        while (videoCapture.read(frame)) {
            int j = 0;
            for (; j < 15 && it != begin; j++) {
                auto point = (*it).getPoint();
                auto hasBounced = (*it).hasBounced();
                it--;
                auto previousPoint = (*it).getPoint();
                it++;
                if (point != NULL_POINT && previousPoint != NULL_POINT) {
                    cv::line(frame, point, previousPoint, Scalar(0, 200, 0), 1);
                    if (hasBounced) {
                        cv::circle(frame, point, 10, Scalar(255, 0, 0), -1);
                    } else {
                        cv::circle(frame, point, 3, Scalar(0, 255, 0), -1);
                    }
                }
                it--;
            }
            for (; j > 0; j--)
                it++;
            videoWriter.write(frame);
            while ((*it).hasBounced())
                it++;
            it++;
        }
        videoWriter.release();
    }
}